/**
 ** WIMP -- WIesbaden MIps Processor
 **
 ** WIMP Assembler
 **
 ** Usage: was <program.asm> [<program.wo>]
 **
 ** Copyright (c) 2014, Robert Kaiser, RheinMain University of Applied Sciences
 ** All rights reserved.
 ** 
 ** Redistribution and use in source and binary forms, with or without modification,
 ** are permitted provided that the following conditions are met:
 **
 ** 1. Redistributions of source code must retain the above copyright notice, this
 **    list of conditions and the following disclaimer.
 ** 2. Redistributions in binary form must reproduce the above copyright notice,
 **    this list of conditions and the following disclaimer in the documentation
 **    and/or other materials provided with the distribution.
 ** 
 ** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 ** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 ** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 ** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 ** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 ** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 ** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 ** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 ** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 ** OF SUCH DAMAGE.
 **
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

/**
 * (mis-)use net functions for optional byte swapping:
 */
#include <netinet/in.h>
#define swapl(x) htonl(x)

#define SIGN_XTND(x) ((int32_t)((int16_t)(x)))

#include "wimp-opcodes.h"

#define NUM_REGISTERS  16

/** flag: set if we need another pass #2 */
int need_more_passes = 0;

/** Additional pseudo-opcodes understood by the assembler */
enum pseudo_opcode
{
    ORG,
    EQU,
    DC,
    DS,
    DB,
    /* synthesized instructions start here */
    MOVE,
    NOP,
    NOT,
    NEG,
    LI,
    LA,
    B,
    BAL,
    BLT,
    BGT,
    BLE,
    BGE,
    BGTU,
    BEQZ,
    SUBI,
    SUBIU,
    LDB,
    STB,
};

/**
 *  Structure defining instructions and their expected parameters
 */
struct instruction
{
    char *Name;                /** Name string */
    enum opcode    Opcode;     /** Opcode */
    enum funcode   Funcode;    /** Function code (type R instructions only)  */
    enum condcode  Cond;       /** Condition code (conditional branch instructions only)  */
    enum pseudo_opcode Pseudo; /** pseudo opcode */
    enum optype { T_NONE, T_REG, T_PTR, T_IMM, T_BIG, T_LBL, T_STR } op[3]; /** expected operands */
};

/** Instruction set */
static struct instruction TheInstructionSet[] =
{
    {"SLL"    , RTYPE , SLL,     0, 0, {T_REG, T_REG,  T_IMM}},
    {"SRL"    , RTYPE , SRL,     0, 0, {T_REG, T_REG,  T_IMM}},
    {"SRA"    , RTYPE , SRA,     0, 0, {T_REG, T_REG,  T_IMM}},
    {"SLLV"   , RTYPE , SLLV,    0, 0, {T_REG, T_REG,  T_REG}},
    {"SRLV"   , RTYPE , SRLV,    0, 0, {T_REG, T_REG,  T_REG}},
    {"SRAV"   , RTYPE , SRAV,    0, 0, {T_REG, T_REG,  T_REG}},
    {"JR"     , RTYPE , JR,      0, 0, {T_REG, T_NONE, T_NONE}},
    {"JALR"   , RTYPE , JALR,    0, 0, {T_REG, T_REG,  T_NONE}},
    {"SYSCALL", RTYPE , SYSCALL, 0, 0, {T_IMM, T_NONE, T_NONE}},
    {"BREAK"  , RTYPE , BREAK,   0, 0, {T_IMM, T_NONE, T_NONE}},
    {"MFHI"   , RTYPE , MFHI,    0, 0, {T_REG, T_NONE, T_NONE}},
    {"MTHI"   , RTYPE , MTHI,    0, 0, {T_REG, T_NONE, T_NONE}},
    {"MFLO"   , RTYPE , MFLO,    0, 0, {T_REG, T_NONE, T_NONE}},
    {"MTLO"   , RTYPE , MTLO,    0, 0, {T_REG, T_NONE, T_NONE}},
    {"MULT"   , RTYPE , MULT,    0, 0, {T_REG, T_REG,  T_NONE}},
    {"MULTU"  , RTYPE , MULTU,   0, 0, {T_REG, T_REG,  T_NONE}},
    {"DIV"    , RTYPE , DIV,     0, 0, {T_REG, T_REG,  T_NONE}},
    {"DIVU"   , RTYPE , DIVU,    0, 0, {T_REG, T_REG,  T_NONE}},
    {"ADD"    , RTYPE , ADD,     0, 0, {T_REG, T_REG,  T_REG}},
    {"ADDU"   , RTYPE , ADDU,    0, 0, {T_REG, T_REG,  T_REG}},
    {"SUB"    , RTYPE , SUB,     0, 0, {T_REG, T_REG,  T_REG}},
    {"SUBU"   , RTYPE , SUBU,    0, 0, {T_REG, T_REG,  T_REG}},
    {"AND"    , RTYPE , AND,     0, 0, {T_REG, T_REG,  T_REG}},
    {"OR"     , RTYPE , OR,      0, 0, {T_REG, T_REG,  T_REG}},
    {"XOR"    , RTYPE , XOR,     0, 0, {T_REG, T_REG,  T_REG}},
    {"NOR"    , RTYPE , NOR,     0, 0, {T_REG, T_REG,  T_REG}},
    {"SLT"    , RTYPE , SLT,     0, 0, {T_REG, T_REG,  T_REG}},
    {"SLTU"   , RTYPE , SLTU,    0, 0, {T_REG, T_REG,  T_REG}},
    {"BLTZ"   , BRANCH, 0,     LTZ, 0, {T_REG, T_LBL,  T_NONE}},
    {"BGEZ"   , BRANCH, 0,     GEZ, 0, {T_REG, T_LBL,  T_NONE}},
    {"BLTZAL" , BRANCH, 0,   LTZAL, 0, {T_REG, T_LBL,  T_NONE}},
    {"BGEZAL" , BRANCH, 0,   GEZAL, 0, {T_REG, T_LBL,  T_NONE}},
    {"J"      , JUMP,   0,       0, 0, {T_BIG, T_NONE, T_NONE}},
    {"JAL"    , JAL,    0,       0, 0, {T_BIG, T_NONE, T_NONE}},
    {"BEQ"    , BEQ,    0,       0, 0, {T_REG, T_REG,  T_LBL}},
    {"BNE"    , BNE,    0,       0, 0, {T_REG, T_REG,  T_LBL}},
    {"BLEZ"   , BLEZ,   0,       0, 0, {T_REG, T_LBL,  T_NONE}},
    {"BGTZ"   , BGTZ,   0,       0, 0, {T_REG, T_LBL,  T_NONE}},
    {"ADDI"   , ADDI,   0,       0, 0, {T_REG, T_REG,  T_IMM}},
    {"ADDIU"  , ADDIU,  0,       0, 0, {T_REG, T_REG,  T_IMM}},
    {"SLTI"   , SLTI,   0,       0, 0, {T_REG, T_REG,  T_IMM}},
    {"SLTIU"  , SLTIU,  0,       0, 0, {T_REG, T_REG,  T_IMM}},
    {"ANDI"   , ANDI,   0,       0, 0, {T_REG, T_REG,  T_IMM}},
    {"ORI"    , ORI,    0,       0, 0, {T_REG, T_REG,  T_IMM}},
    {"XORI"   , XORI,   0,       0, 0, {T_REG, T_REG,  T_IMM}},
    {"LUI"    , LUI,    0,       0, 0, {T_REG, T_IMM,  T_NONE}},

//     {"LB"     , LB,     0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"LH"     , LH,     0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"LWL"    , LWL,    0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"LW"     , LW,     0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"LBU"    , LBU,    0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"LHU"    , LHU,    0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"LWR"    , LWR,    0,       0, 0, {T_REG, T_IMM,  T_PTR}},
//    {"SB"     , SB,     0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"SH"     , SH,     0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"SWL"    , SWL,    0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"SW"     , SW,     0,       0, 0, {T_REG, T_IMM,  T_PTR}},
    {"SWR"    , SWR,    0,       0, 0, {T_REG, T_IMM,  T_PTR}},

    /** Pseudo instructions: */
    {"ORG"    , PSEUDO, 0,       0, ORG,{T_BIG,  T_NONE, T_NONE}},
    {"EQU"    , PSEUDO, 0,       0, EQU,{T_BIG,  T_NONE, T_NONE}},
    {"DC"     , PSEUDO, 0,       0, DC, {T_BIG,  T_NONE, T_NONE}},
    {"DS"     , PSEUDO, 0,       0, DS, {T_BIG,  T_NONE, T_NONE}},
    {"DB"     , PSEUDO, 0,       0, DB, {T_STR,  T_NONE, T_NONE}},

    /** Synthesized instructions: */
    {"MOVE"   , SYNTH,  0,       0, MOVE,{T_REG,  T_REG,  T_NONE}},
    {"NOP"    , SYNTH,  0,       0, NOP, {T_NONE, T_NONE, T_NONE}},
    {"NOT"    , SYNTH,  0,       0, NOT, {T_REG,  T_REG,  T_NONE}},
    {"NEG"    , SYNTH,  0,       0, NEG, {T_REG,  T_NONE, T_NONE}},
    {"LI"     , SYNTH,  0,       0, LI,  {T_REG,  T_BIG,  T_NONE}},
    {"LA"     , SYNTH,  0,       0, LA,  {T_REG,  T_BIG,  T_NONE}},
    {"B"      , SYNTH,  0,       0, B,   {T_LBL,  T_NONE, T_NONE}},
    {"BAL"    , SYNTH,  0,       0, BAL, {T_LBL,  T_NONE, T_NONE}},
    {"BLT"    , SYNTH,  0,       0, BLT, {T_REG,  T_REG,  T_LBL }},
    {"BGT"    , SYNTH,  0,       0, BGT, {T_REG,  T_REG,  T_LBL }},
    {"BLE"    , SYNTH,  0,       0, BLE, {T_REG,  T_REG,  T_LBL }},
    {"BGE"    , SYNTH,  0,       0, BGE, {T_REG,  T_REG,  T_LBL }},
    {"BGTU"   , SYNTH,  0,       0, BGTU,{T_REG,  T_REG,  T_LBL }},
    {"BEQZ"   , SYNTH,  0,       0, BEQZ,{T_REG,  T_LBL,  T_NONE}},
    {"SUBI"   , SYNTH,  0,       0, SUBI,{T_REG,  T_REG,  T_IMM }},
    {"SUBIU"  , SYNTH,  0,       0, SUBIU,{T_REG, T_REG,  T_IMM }},
    
    {"LB"     , SYNTH,  0,       0, LDB, {T_REG,  T_IMM,  T_PTR}},
    {"SB"     , SYNTH,  0,       0, STB, {T_REG,  T_IMM,  T_PTR}},
};

#define NUM_OPCODES sizeof(TheInstructionSet)/sizeof(struct instruction)

struct regname
{
    char *Name;          /** Name string */
    int  num;            /** Register number */
};

static struct regname RegisterNames[] =
{
    {"ZERO", 0 }, {"AT",   1 }, {"V0",   2 }, {"V1",   3 },
    {"A0",   4 }, {"A1",   5 }, {"A2",   6 }, {"A3",   7 },
    {"T0",   8 }, {"T1",   9 }, {"T2",   10}, {"T3",   11},
    {"T4",   12}, {"T5",   13}, {"T6",   14}, {"T7",   15},
    {"S0",   16}, {"S1",   17}, {"S2",   18}, {"S3",   19},
    {"S4",   20}, {"S5",   21}, {"S6",   22}, {"S7",   23},
    {"T8",   24}, {"T9",   25}, {"K0",   26}, {"K1",   27},
    {"GP",   28}, {"SP",   29}, {"FP",   30}, {"S8",   30}, {"RA",   31},
    {"R0", 0 },   {"R1",   1 }, {"R2",   2 }, {"R3",   3 },
    {"R4",   4 }, {"R5",   5 }, {"R6",   6 }, {"R7",   7 },
    {"R8",   8 }, {"R9",   9 }, {"R10",  10}, {"R11",  11},
    {"R12",  12}, {"R13",  13}, {"R14",  14}, {"R15",  15},
    {"R16",  16}, {"R17",  17}, {"R18",  18}, {"R19",  19},
    {"R20",  20}, {"R21",  21}, {"R22",  22}, {"R23",  23},
    {"R24",  24}, {"R25",  25}, {"R25",  26}, {"R27",  27},
    {"R28",  28}, {"R29",  29}, {"R30",  30}, {"R31",  31},
};

#define NUM_REGNAMES sizeof(RegisterNames)/sizeof(struct regname)

/** Symbol table entry: */
struct symrec
{
       char *name;  /** name of symbol              */
       int value;   /** associated value */
       int equ;    /** value assigned throug EQU */
       struct symrec *next;    /** link field    */
};

typedef struct symrec symrec;

/** The symbol table: a chain of `struct symrec'.  */
symrec *sym_table = (symrec *)0;

/** append an entry to the symbol table */
static symrec *putsym(char *sym_name, int sym_val)
{
       symrec *ptr;
       ptr = (symrec *) malloc (sizeof(symrec));
       ptr->name = (char *) malloc (strlen(sym_name)+1);
       strcpy (ptr->name,sym_name);
       ptr->value = sym_val;
       ptr->next = (struct symrec *)sym_table;
       ptr->equ = 0;
       sym_table = ptr;
       return ptr;
}

/** lookup a symbol by name */
static symrec *getsym(char *sym_name)
{
       symrec *ptr;
       for (ptr = sym_table; ptr != (symrec *) 0;
            ptr = (symrec *)ptr->next)
         if (strcmp (ptr->name,sym_name) == 0)
           return ptr;
       return NULL;
}

/** change the value associated to the most recently created entry of the symbol table */
static symrec *setsym(char* name, int old_val, int new_val, int pass)
{
    symrec *ptr;
    if(pass == 1)
    {
        if(sym_table->value != old_val)
            return NULL;
        sym_table->value = new_val;
        sym_table->equ = 1;
        return(sym_table);
    }
    else
    {   /** pass = 2: label name must be in table already */
       for (ptr = sym_table; ptr != (symrec *) 0;  ptr = (symrec *)ptr->next)
       {
         if (!strcmp(ptr->name, name))
         {
             ptr->value = new_val;
             return ptr;
         }
       }
       return NULL;
    }
}

static void init_table (void)  /** puts any predefined constants in table. */
{
       symrec *ptr __attribute__((unused));
       ptr = putsym("_start", 0);
}

static struct instruction *parse_instruction(char *s)
{
    int i, len;
    struct instruction *instr;
    for(i = 0; i < NUM_OPCODES; i++)
    {
        instr = &TheInstructionSet[i];
        len = strlen(instr->Name);
        if(!strncasecmp(s, instr->Name, len) && (s[len] == '\0' || s[len] == ' ' || s[len] == '\t'))
            return instr;
    }
    return NULL;
}

static char* parse_reg(char *s, int *num)
{
    int regnum, i, len;
    char *sep;
    struct regname *reg;

    for(sep = s; *sep != '\0' && *sep != ',' && *sep != ' ' && *sep != '\t'; ++sep)
        ;
    if(*sep != '\0')
        *sep++ = '\0';

    if(*s == '$') /** registers may be denoted as $<number> ... */
    {
        ++s;
        if(sscanf(s, "%d", &regnum) == 1)
        {
            if(regnum >= 0 && regnum < NUM_REGISTERS)
            {
                *num = regnum;
                return sep;
            }
        }
        else /** also allow "$a0" etc. */
        {
            for(i = 0; i < NUM_REGNAMES; i++)
            {
                reg = &RegisterNames[i];
                len = strlen(reg->Name);
                if(!strncasecmp(s, reg->Name, len) && (s[len] == '\0' || s[len] == ' ' || s[len] == '\t' || s[len] == ')'))
                {
                    *num = reg->num;
                    return sep;
                }
            }
        }
    }
    else /** ... or by name: */
    {
        for(i = 0; i < NUM_REGNAMES; i++)
        {
            reg = &RegisterNames[i];
            len = strlen(reg->Name);
            if(!strncasecmp(s, reg->Name, len) && (s[len] == '\0' || s[len] == ' ' || s[len] == '\t' || s[len] == ')'))
            {
                *num = reg->num;
                return sep;
            }
        }
    }
    return NULL;
}

static char *parse_num(char *s, int *num, int *val_known, int pass)
{
    symrec *sym;
    int number;
    char *sep, *end;

    /**
     * step forward to find end of number/symbol name
     * Note: since we may be parsing an offset, '(' is a valid end marker
     */
    for(sep = s; *sep != '\0' && *sep != ',' && *sep != ' ' && *sep != '\t' && *sep != '('; ++sep)
        ;

    if(*sep != '\0' && *sep != '(') /* terminate the string */
        *sep++ = '\0';

    number = strtoul(s, &end, 0);
    if(end != s)
    { /** if it is a valid number, return its value */
        *num = number;
        *val_known = 1;
        return sep;
    }
    else /** try to find value of symbol in table */
    {
        char save = *sep;
        *sep = '\0';
        if((sym = getsym(s)) != NULL)
        {
            *num = sym->value;
            *sep = save;
            *val_known = 1;
            return sep;
        }
        else if(pass == 1) /** first pass: just accept non-numbers */
        {
            *val_known = 0;
            return sep;
        }
        else /** second pass: must find value of symbol in table */
        {
             printf("Unknown symbol: '%s' ", s);
        }
    }
    *val_known = 1;
    return NULL;
}

static char *parse_string(char *s, int *len)
{
    int number;
    char *sep, *stor, *ret;

    for(sep = s; *sep != '\0' && *sep != '"'; ++sep)
        ;
    if(*sep == '\0')
        return(NULL);
    else
        ++sep;

    ret = stor = sep;
    for(number = 0; *sep != '"' && *sep != '\0'; number++)
    {
        if(*sep == '\\')
        { /** escape char: replace it by actual code */
            switch(*(sep + 1))
            {
            case '\0':
                return NULL; /** error: unterminated string */
                break;
            case 'n':
                *stor++ = '\n';
                sep += 2;
                break;
            case 'r':
                *stor++ = '\r';
                sep += 2;
                break;
            case 't':
                *stor++ = '\t';
                sep += 2;
                break;
            case 'b':
                *stor++ = '\b';
                sep += 2;
                break;
            case '\\':
                *stor++ = '\\';
                sep += 2;
                break;
            case '"':
                *stor++ = '"';
                sep += 2;
                break;
            default:
                *stor++ = *(sep + 1);
                sep += 2;
                break;
            }
        }
        else
        {
            *stor++ = *sep++;
        }
    }
    if(*sep == '"')
    {
        *stor++ = '\0';
        *len = number;
        return ret;
    }
    return NULL;
}


static char *extract_label(char*s)
{
    char *sep, *blank, *tab;;

    sep   = strchr(s, ':');
    blank = strchr(s, ' ');
    tab   = strchr(s, '\t');

    if(sep) /** ":" in text */
    {   /** Check if ":" is preceeded by " " or "\t" */
        if(blank && sep > blank)
            return NULL; 
        else if(tab && sep > tab)
            return NULL;
        else
            return sep; /** sep points to ":" */
    }
    return NULL;
}

static int bra_dist(int target, int currpos)
{
    return((target - currpos - 4)/4);
}

static int emit_string(char*sep, int num, int pass, FILE *out)
{
    int len, advance = 0;
    while(num >= sizeof(int))
    {
        if(pass == 2) /** write operand to output file */
        {
            fwrite(sep, sizeof(int), 1, out);
        }
        num -= sizeof(int);
        sep += sizeof(int);
        advance += sizeof(int);
    }
    len = sizeof(int) - num;
    while(num > 0)
    {
        if(pass == 2) /** write operand to output file */
        {
            fwrite(sep, 1, 1, out);
        }
        num--;
        sep++;
    }
    while(len > 0)
    {
        if(pass == 2) /** fill remainder with zeroes */
        {
            fputc(0, out);
        }
        len--;
    }
    advance += sizeof(int);
    return advance;
}

static WimpInst_t make_opcode(struct instruction *instr, int ops[3], int location, int pass, FILE *out)
{
    WimpInst_t I, uCode[2];
    int count = 0, i;

    switch(instr->Opcode)
    {
      case RTYPE:
        switch(instr->Funcode)
        {
          case SLL:
          case SRL:
          case SRA:
              uCode[count++] = RTYPE(instr->Opcode, 0, ops[1], ops[0], ops[2], instr->Funcode);
            break;
          case SLLV:
          case SRLV:
          case SRAV:
            uCode[count++] = RTYPE(instr->Opcode, ops[2], ops[1], ops[0], 0, instr->Funcode);
            break;
          case JR:
            uCode[count++] = RTYPE(instr->Opcode, ops[0], 0, 0, 0, instr->Funcode);
            break;
          case JALR: /** emit a NOP after JALR to ensure MIPS branch delay compatibility */
            uCode[count++] = RTYPE(instr->Opcode, ops[1], 0, ops[0], 0, instr->Funcode);
            uCode[count++] = RTYPE(RTYPE, 0, 0, 0, 0, SLL);
            break;
          case SYSCALL:
          case BREAK:
            uCode[count++] = RTRAP(instr->Opcode, ops[0], instr->Funcode);
            break;
          case MFHI:
          case MFLO:
            uCode[count++] = RTYPE(instr->Opcode, 0, 0, ops[0], 0, instr->Funcode);
            break;
          case MTHI:
          case MTLO:
            uCode[count++] = RTYPE(instr->Opcode, ops[0], 0, 0, 0, instr->Funcode);
            break;
          case MULT:
          case MULTU:
          case DIV:
          case DIVU:
            uCode[count++] = RTYPE(instr->Opcode, ops[0], ops[1], 0, 0, instr->Funcode);
            break;
          case ADD:
          case ADDU:
          case SUB:
          case SUBU:
          case AND:
          case OR:
          case XOR:
          case NOR:
          case SLT:
          case SLTU:
            uCode[count++] = RTYPE(instr->Opcode, ops[1], ops[2], ops[0], 0, instr->Funcode);
            break;
        default:
            assert(0); /* should never get here! */
            break;
        }
        break;
      case BRANCH:
        uCode[count++] = ITYPE(instr->Opcode, ops[0], instr->Cond, bra_dist(ops[1], location));
        break;
      case JUMP:
        uCode[count++] = JTYPE(instr->Opcode, ops[0] >> 2);
        break;
      case JAL:    /** emit a NOP after JAL to ensure MIPS branch delay compatibility */
        uCode[count++] = JTYPE(instr->Opcode, ops[0] >> 2);
        uCode[count++] = RTYPE(RTYPE, 0, 0, 0, 0, SLL);
        break;
      case BEQ:
      case BNE:
        uCode[count++] = ITYPE(instr->Opcode, ops[0], ops[1], bra_dist(ops[2], location));
        break;
      case BLEZ:
      case BGTZ:
        uCode[count++] = ITYPE(instr->Opcode, ops[0], 0, bra_dist(ops[1], location));
        break;
      case ADDI:
      case ADDIU:
      case SLTI:
      case SLTIU:
      case ANDI:
      case ORI:
      case XORI:
        uCode[count++] = ITYPE(instr->Opcode, ops[1], ops[0], ops[2]);
        break;
      case LUI:
        uCode[count++] = ITYPE(instr->Opcode, 0, ops[0], ops[1]);
        break;
      case LB:
      case LH:
      case LWL:
      case LW:
      case LBU:
      case LHU:
      case LWR:
      case SB:
      case SH:
      case SWL:
      case SWR:
      case SW:
        uCode[count++] = ITYPE(instr->Opcode, ops[2], ops[0], ops[1]);
        break;
      default:
        assert(0); /* should never get here! */
        break;
    }
    assert(count <= 2);
    if(pass == 2) /**Emit the code */
    {
        for(i = 0; i < count; i++)
        {
            I = swapl(uCode[i]);
            fwrite(&I, sizeof I, 1, out);
        }
    }
    return count * sizeof I;
}

static int synth_opcodes(struct instruction *instr, int ops[3], int location, int pass, FILE *out)
{
    WimpInst_t I, uCode[30];
    int count = 0, i;
    assert(instr->Opcode == SYNTH);
    switch(instr->Pseudo)
    {
        case MOVE:
            /** synthesize as addu rt,rs,$0 */
            uCode[count++] = RTYPE(RTYPE, ops[1], 0, ops[0], 0, ADDU);
            break;
        case NOP:
            /** synthesize as sll $0,$0,0 */
            uCode[count++] = RTYPE(RTYPE, 0, 0, 0, 0, SLL);
            break;
        case NOT:
            /** synthesize as nor $rt, $rs, $zero */
            uCode[count++] = RTYPE(RTYPE, ops[1], 0, ops[0], 0, NOR);
            break;
        case NEG: /** negate: flip all bits (-> one's complement), then add 1 (two's complement) */
            /** synthesize as nor $rt, $rt, $zero; addiu $rt, $rt, 1 */
            uCode[count++] = RTYPE(RTYPE, ops[0], 0, ops[0], 0, NOR);
            uCode[count++] = ITYPE(ADDIU, ops[0], 1, ops[0]);
            break;
        case LI:
        case LA:
            if(SIGN_XTND(ops[1] & 0xffff) == ops[1]) /** is immediate small enough ? */
            {    /** synthesize as addiu rt,$0,imm */
                uCode[count++] = ITYPE(ADDIU, 0, ops[0], ops[1]);
            }
            else if((ops[1] & 0xffff) == 0) /** is lower halfword = 0 ? */
            {    /** synthesize as lui $rt,imm >> 16 */
                uCode[count++] = ITYPE(LUI,   0, ops[0], ops[1] >> 16);
            }
            else
            {    /** synthesize as lui $rt,imm >> 16; ori $rt,$rt, imm & 0xffff */
                uCode[count++] = ITYPE(LUI,   0, ops[0], ops[1] >> 16);
                uCode[count++] = ITYPE(ORI, ops[0], ops[0], ops[1] & 0xffff);
            }
            break;
        case B:
            /** synthesize as beq $zero,$zero,Label */
            uCode[count++] = ITYPE(BEQ, 0, 0, bra_dist(ops[0], location));
            break;
        case BAL:
            /** synthesize as  bgezal $zero,Label */
            uCode[count++] = ITYPE(BRANCH, 0, GEZAL, bra_dist(ops[0], location));
            break;
        case BLT:
            /** synthesize as slt $at,$rs,$rt; bne $at,$zero,Label */
            uCode[count++] = RTYPE(RTYPE, ops[0], ops[1], 1, 0, SLT);
            uCode[count++] = ITYPE(BNE, 1, 0, bra_dist(ops[2], location + sizeof I));
            break;
        case BGT:
            /** synthesize as slt $at,$rt,$rs; bne $at,$zero,Label */
            uCode[count++] = RTYPE(RTYPE, ops[1], ops[0], 1, 0, SLT);
            uCode[count++] = ITYPE(BNE, 1, 0, bra_dist(ops[2], location + sizeof I));
            break;
        case BLE:
            /** synthesize as slt $at,$rt,$rs; beq $at,$zero,Label */
            uCode[count++] = RTYPE(RTYPE, ops[1], ops[0], 1, 0, SLT);
            uCode[count++] = ITYPE(BEQ, 1, 0, bra_dist(ops[2], location + sizeof I));
            break;
        case BGE:
            /** synthesize as slt $at,$rs,$rt; beq $at,$zero,Label */
            uCode[count++] = RTYPE(RTYPE, ops[0], ops[1], 1, 0, SLT);
            uCode[count++] = ITYPE(BEQ, 1, 0, bra_dist(ops[2], location + sizeof I));
            break;
        case BGTU:
            /** synthesize as sltu $at,$rt,$rs; bne $at,$zero,Label */
            uCode[count++] = RTYPE(RTYPE, ops[1], ops[0], 1, 0, SLTU);
            uCode[count++] = ITYPE(BNE, 1, 0, bra_dist(ops[2], location + sizeof I));
            break;
        case BEQZ:
            /** synthesize as beq $rs,$zero,Label */
            uCode[count++] = ITYPE(BEQ, ops[0], 0, bra_dist(ops[1], location));
            break;
        case SUBI:
            /** synthesize as addi $rs,$rt,-immediate */
            uCode[count++] = ITYPE(ADDI, ops[1], ops[0], -ops[2]);
            break;
        case SUBIU:
            /** synthesize as addiu $rs,$rt,-immediate */
            uCode[count++] = ITYPE(ADDIU, ops[1], ops[0], -ops[2]);
            break;
        case LDB:
           /** lb rt,offset(base): synthesize as:
             *  addiu at,base,offset ; get effective address
             *  andi  at,at,0xfffc   ; make it word aligned
             *  lw    rt,0(at)       ; load memory word
             *  addiu at,base,offset ; get effective address once more
             *  andi  at,at,0x3      ; isolate byte number (0..3)
             *  addiu at,at,-3        ; make that (-3..0)
             *  nor   at,at,$zero    ; negate: step 1: 1's complement
             *  addiu at,at,1        ; negate: step 2: add 1 to get 2's complement
             *  sll   at,at,3        ; times 8 to get shift amount
             *  srlv  rt,rt,at       ; shift requested byte into right position
             *  andi  rt,rt,0xff     ; clear high bits
             *
             * NOTE: This is AWFUL, I know ...
             * But supporting non-word, misaligned transfers would have made the
             * machine architecture way more complex. Actually, the compiler
             * is expected to handle this, so the assembler will never see
             * lb instructions in compiler-generated code. This crappy approximation
             * is only provided for the poor souls who program this machine
             * in plain assembler
             */
            uCode[count++] = ITYPE(ADDIU, ops[2], 1, ops[1]);
            uCode[count++] = ITYPE(ANDI, 1, 1, 0xfffc);
            uCode[count++] = ITYPE(LW, 1, ops[0], 0);
            uCode[count++] = ITYPE(ADDIU, ops[2], 1, ops[1]);
            uCode[count++] = ITYPE(ANDI, 1, 1, 3);
            uCode[count++] = ITYPE(ADDIU, 1, 1, -3);
            uCode[count++] = RTYPE(RTYPE, 1, 0, 1, 0, NOR);
            uCode[count++] = ITYPE(ADDIU, 1, 1, 1);
            uCode[count++] = RTYPE(RTYPE, 0, 1, 1, 3, SLL);
            uCode[count++] = RTYPE(RTYPE, 1, ops[0], ops[0], 0, SRLV);
            uCode[count++] = ITYPE(ANDI, ops[0], ops[0], 0xff);
            break;
        case STB:
        {
           /** sb rs,offset(base): synthesize as:
             *  addiu sp,sp,-12        ; get stack space for 3xword on stack
             *  sw    tmp0,0(sp)       ; save registers
             *  sw    tmp1,4(sp)
             *  sw    rs,8(sp)
             *  addiu tmp0,base,offset ; get effective address
             *  andi  tmp1,tmp0,0xfffc ; make it word aligned
             *  andi  tmp0,tmp0,0x3    ; isolate byte number (0..3)
             *  addiu tmp0,tmp0,-3     ; make that (-3..0)
             *  nor   tmp0,tmp0,$zero  ; negate: step 1: 1's complement
             *  addiu tmp0,tmp0,1      ; negate: step 2: add 1 to get 2's complement
             *  sll   tmp0,tmp0,3      ; times 8 to get shift amount
             *  addiu at,zero,0xff     ; get byte mask
             *  sllv  at,at,tmp0       ; shift it into right position
             *  sllv  rs,rs,tmp0       ; also shift byte to be written
             *  and   rs,rs,at         ; mask out
             *  nor   at,at,$zero      ; negate mask
             *  lw    tmp0,0(tmp1)
             *  and   tmp0,tmp0,at
             *  or    rs,rs,tmp0
             *  sw    rs,0(tmp1)
             *  lw    tmp0,0(sp)       ; restore Registers
             *  lw    tmp1,4(sp)
             *  lw    rs,8(sp)
             *  addiu sp,sp,12         ; release stack space 
             *
             * NOTE: This is EVEN MORE AWFUL, I know ...
             * (See above for some rationale). This code requires two
             * more scratch registers to spill temporarily. We need
             * to choose them so they do not overlap with the other used
             * registers, i.e. the chosen regs must not be any of
             * $zero (=$0), $AT (=$1), $k0 (=$26), $k1 (=$27),
             * $sp (=$29), $ra (=$31), rs (=ops[0]) and base (=ops[2]):
             * 
             */
            int tmpreg[2];
            int i, j = 0;
            /** compute the set of available registers. This is
                represented as a bit mask: */
            uint32_t avail_regs = ~((1<<0)|(1<<1)|(1<<26)|(1<<27)|(1<<29)|(1<<31));
            avail_regs &= ~((1<<ops[1])|(1<<ops[2]));
            /** loop for all registers, find two which are available : */
            for(i = 0; i < 32; i++)
            {
                if(!!(avail_regs & (1<<i))) /** register #i available? */
                {
                    tmpreg[j++] = i; /** yes -> save its number */
                    if(j == 2) /** do we have two registers yet? */
                        break; /** yes -> proceed */
                }
            }
            /** now generate the code, using the two registers */
            uCode[count++] = ITYPE(ADDIU, 29, 29, -12);
            uCode[count++] = ITYPE(SW, 29, tmpreg[0], 0);
            uCode[count++] = ITYPE(SW, 29, tmpreg[1], 4);
            uCode[count++] = ITYPE(SW, 29, ops[0], 8);
            uCode[count++] = ITYPE(ADDIU, ops[2], tmpreg[0], ops[1]);
            uCode[count++] = ITYPE(ANDI, tmpreg[0], tmpreg[1], 0xfffc);
            uCode[count++] = ITYPE(ANDI, tmpreg[0], tmpreg[0], 3);
            uCode[count++] = ITYPE(ADDIU, tmpreg[0], tmpreg[0], -3);
            uCode[count++] = RTYPE(RTYPE, tmpreg[0], 0, tmpreg[0], 0, NOR);
            uCode[count++] = ITYPE(ADDIU, tmpreg[0], tmpreg[0], 1);
            uCode[count++] = RTYPE(RTYPE, 0, tmpreg[0], tmpreg[0], 3, SLL);
            uCode[count++] = ITYPE(ADDIU, 0, 1, 0xff);
            uCode[count++] = RTYPE(RTYPE, tmpreg[0], 1, 1, 0, SLLV);
            uCode[count++] = RTYPE(RTYPE, tmpreg[0], ops[0], ops[0], 0, SLLV);
            uCode[count++] = RTYPE(RTYPE, ops[0], 1, ops[0], 0, AND);
            uCode[count++] = RTYPE(RTYPE, 1, 0, 1, 0, NOR);
            uCode[count++] = ITYPE(LW, tmpreg[1], tmpreg[0], 0);
            uCode[count++] = RTYPE(RTYPE, tmpreg[0], 1, tmpreg[0], 0, AND);
            uCode[count++] = RTYPE(RTYPE, ops[0], tmpreg[0], ops[0], 0, OR);
            uCode[count++] = ITYPE(SW, tmpreg[1], ops[0], 0);
            uCode[count++] = ITYPE(LW, 29, tmpreg[0], 0);
            uCode[count++] = ITYPE(LW, 29, tmpreg[1], 4);
            uCode[count++] = ITYPE(LW, 29, ops[0], 8);
            uCode[count++] = ITYPE(ADDIU, 29, 29, 12);
            break;
        }
        default:
            assert(0);
            break;
    }

    if(pass == 2) /**Emit the code */
    {
        for(i = 0; i < count; i++)
        {
            I = swapl(uCode[i]);
            fwrite(&I, sizeof I, 1, out);
        }
    }
    return count * sizeof I;
}


static int assemble(FILE *in, FILE* out, int pass)
{
    int location = 0;
    int linecount = 0;
    int errorcount = 0;
    int len, i, num, val_known, stringlen;
    char *s, *sep, *string, *lastlabel;
    struct instruction *instr;
    symrec *sym;
    int operands[3];
    char linebuffer[512];
    /** read line by line */
    while(fgets(linebuffer, 512, in) != NULL)
    {
        lastlabel = NULL;
        ++linecount;
        /** Skip empty lines */
        if((len = strlen(linebuffer)) == 0)
            continue;
        /** Delete linefeed at end of line */
        if(linebuffer[len-1] == '\n')
        {
            linebuffer[len-1] = '\0';
            if(--len == 0)
               continue;
        }
        /** Skip leading blanks */
        for(s = linebuffer; *s == ' ' || *s == '\t'; s++)
            ;
        if((sep = strchr(s, ';')))
        {
            *sep = '\0'; /** Separate comment (begins with ";") from rest */
        }
        if((sep = strchr(s, '#')))
        {
            *sep = '\0'; /** also accept "#" as comment starter */
        }
        if((len = strlen(s)) == 0)
            continue; /** Nothing left -> next line */

        /** Does line start with a label? */
        if((sep = extract_label(s)))
        {
            *sep = '\0'; /** Separate label from rest (overwrite ":") */
            if(pass == 1)
                putsym(s, location); /** put label into symbol table */
            else
            {   /** second pass: save for setsym() */
                lastlabel = s;
                if((sym = getsym(s)) == NULL)
                { /** must find symbol in symbol table during second pass */
                    printf("Line %d: symbol \'%s\' not found during second pass!\n", linecount, s);
                    ++errorcount;
                    continue;
                }
                else if(sym->value != location && !sym->equ)
                { /**
                   * label value has changed: must repeat pass #2
                   * 
                   * This can happen, for exampe if there is. e.g.
                   * an li <register>,immediate instruction: In pass #1
                   * we may not yet know the value of the immediate and
                   * we assume (optimistically) that it will fit into
                   * 16 bits, so we will generate a single instruction.
                   * However, if during pass #2 we find out that it is
                   * too big, we must generate a two-instruction sequence,
                   * shifting all subsequent labels up.
                   **/
                
                    sym->value = location;
                    need_more_passes = 1;
                    continue;
                }
            }
            /** continue after ":"  */
            /** Skip leading blanks */
            for(s = sep + 1; *s == ' ' || *s == '\t'; s++)
                ;
            /** if remainder of line is empty: read next line */
            if((len = strlen(s)) == 0)
                continue;
        }

        /** After all this fuss, the assembly language instruction is now ready
            for parsing: Look it up in table */
        if((instr = parse_instruction(s)) == NULL)
        {
            printf("Line %d: Unknown assembly instruction!\n", linecount);
            ++errorcount;
            continue;
        }
        /** point to first operand */
        for(s += strlen(instr->Name); *s == ' ' || *s == '\t'; ++s)
            ;

        /** collect operands  */
        for(i = 0; i < 3; i++)
        {
            /** Skip possible blanks */
            for(; *s == ' ' || *s == '\t'; ++s)
                ;
            num = 0;
            switch(instr->op[i])
            {
                case T_NONE:
                    if(*s != '\0')
                    {  /** Error: */
                        printf("Line %d: unexpected operand!\n", linecount);
                        ++errorcount;
                    }
                    operands[i] = 0;
                    break;
                case T_BIG:
                case T_IMM:
                case T_LBL:
                    if((sep = parse_num(s, &num, &val_known, pass)) == NULL)
                    {  /** Error: */
                        printf("Line %d: operand #%d unsuitable!\n", linecount, i+1);
                        ++errorcount;
                        *s = '\0';
                    }
                    else
                    {
                        operands[i] = num;
                        if(val_known)
                        {
                            if(instr->op[i] == T_IMM &&  SIGN_XTND(num & 0xffff) != num)
                            {  /** Error: this number won't fit into 16 bits */
                                printf("Line %d: operand #%d out of range!\n", linecount, i+1);
                                ++errorcount;
                            }
                            if(instr->op[i] == T_LBL &&  SIGN_XTND(bra_dist(num, location) & 0xffff) != bra_dist(num, location))
                            {  /** Error: this branch distance won't fit into 16 bits */
                                printf("Line %d: operand #%d: jump target out of range!\n", linecount, i+1);
                                ++errorcount;
                            }
                        }
                        /** point to next operand */
                        s = sep;
                    }
                    break;
                case T_PTR:
                    if(*s != '(')
                    {  /** Error: */
                        printf("Line %d: \'(\' expected !\n", linecount);
                        ++errorcount;
                        *s = '\0';
                        break;
                    }
                    else
                    {
                        ++s; /* skip '(' */
                    }
                    /* Fall through to parse the register */
                case T_REG:
                    if((sep = parse_reg(s, &num)) == NULL)
                    {  /** Error: */
                        printf("Line %d: operand #%d unsuitable !\n", linecount, i+1);
                        ++errorcount;
                        *s = '\0';
                    }
                    else
                    {
                        operands[i] = num;
                        /** point to next operand */
                        s = sep;
                    }
                    break;
                case T_STR:
                    if((sep = parse_string(s, &num)) == NULL)
                    {  /** Error: */
                        printf("Line %d: Bad string !\n", linecount);
                        ++errorcount;
                        *s = '\0';
                    }
                    else  /** Bytestring: need special treatment */
                    {
                        string = sep;
                        stringlen = num;
                        *s = '\0';
                    }
                    break;
                 default:
                    printf("Line %d: internal error!\n", linecount);
                    exit(1);
                    break;
            }
       }
       if(instr->Opcode == PSEUDO)
       {    /** pseudo instructions */
           switch(instr->Pseudo)
           {
               case DS:
               {  /** DS = Define Storage */
                   int j, null = 0;
                   for(j = 0; j < operands[0]; j++)
                   {
                       if(pass == 2) /** write operand to output file */
                           fwrite(&null, sizeof null, 1, out);
                       location += sizeof(int);
                   }
                   break;
               }
               case DC:
               {   /** DC = Define Constant */
                   WimpInst_t I = swapl(operands[0]);
                   if(pass == 2) /** Just emit the value */
                       fwrite(&I, sizeof I, 1, out);
                   location += sizeof I;
                   break;
               }
               case DB:
               { /** Bytestring: need special treatment */
                   location += emit_string(string, stringlen, pass, out);
                   *s = '\0';
                   break;
               }
               case ORG:
               {
                   int j, words, null = 0;
                   if(operands[0] % 4 != 0)
                   {
                        printf("Line %d: Origin must be aligned !\n", linecount);
                        ++errorcount;
                   }
                   else if(operands[0] < location)
                   {
                        printf("Line %d: Cant go backwards with ORG !\n", linecount);
                        ++errorcount;
                   }
                   else if(location > 0) /** Already emitted something ? */
                   {
                       if(pass == 2) /** dump zeroes to output file */
                       {
                           words = (operands[0] - location) / sizeof null;
                           for(j = 0; j < words; j++)
                           {
                               fwrite(&null, sizeof null, 1, out);
                           }
                       }
                   }
                   location = operands[0];
                   break;
               }
               case EQU:
               {
                   if(!setsym(lastlabel, location, operands[0], pass))
                   {
                        printf(pass == 1 ? "Line %d: EQU must be preceeded by a label !\n": "Line %d: symbol not resolved !\n" , linecount);
                        ++errorcount;
                   }
                   break;
               }
               default:
                   assert(0); /* should never get here! */
                   break;
            }
       }
       else if(instr->Opcode == SYNTH)
       {    /** Synthesized instruction */
           location += synth_opcodes(instr, operands, location, pass, out);
       }
       else /** emit opcode, count location */
       {
           location += make_opcode(instr, operands, location, pass, out);
       }
    }
    return errorcount;
}


int main (int argc, char *argv[])
{
  char *infile, *outfile;
  FILE *in, *out;
  if (argc < 2)
  {
      printf ("Usage: %s <input file> [<output file>]\n", argv[0]);
      exit(1);
  }
  else
  {
      infile = argv[1];
  }
  if (argc == 3)
      outfile = argv[2];
  else
      outfile = "out.wo";

  in = fopen (infile, "r");

  if(strcmp(outfile, "-"))
      out = fopen (outfile, "w");
  else
      out = stdout;
  if(in == NULL)
  {
    printf("Cannot open file '%s'\n", infile);
    exit(1);
  }
  if(out == NULL)
  {
    printf("Cannot create file '%s'\n", outfile);
    exit(1);
  }

  init_table();

  /** Pass #1: Build symbol table */
  if(assemble(in, NULL, 1) != 0)
  {
      printf("Error during assembly\n");
      exit(1);
  }

  do
  {
      need_more_passes = 0;
      rewind(in);
      rewind(out);

      /** Pass #2: generate code */
      if(assemble(in, out, 2) != 0)
      {
          printf("Assembly error in Pass #2\n");
          exit(1);
      }
  }
  while(need_more_passes);
  return(0);
}
