/**
 ** WIMP -- WIesbaden MIps Processor
 **
 ** WImp disassembler
 **
 ** Usage: wid <program.wo>
 **
 ** Copyright (c) 2014, Robert Kaiser, RheinMain University of Applied Sciences
 ** All rights reserved.
 ** 
 ** Redistribution and use in source and binary forms, with or without modification,
 ** are permitted provided that the following conditions are met:
 **
 ** 1. Redistributions of source code must retain the above copyright notice, this
 **    list of conditions and the following disclaimer.
 ** 2. Redistributions in binary form must reproduce the above copyright notice,
 **    this list of conditions and the following disclaimer in the documentation
 **    and/or other materials provided with the distribution.
 ** 
 ** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 ** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 ** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 ** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 ** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 ** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 ** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 ** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 ** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 ** OF SUCH DAMAGE.
 **
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>


#include "wimp-opcodes.h"
#include "wimp-dis.h"

#define NUM_REGISTERS  32

#define SIGN_XTND(x) ((int32_t)((int16_t)(x)))


static char *reg[NUM_REGISTERS] = {
    "zero", "at", "v0", "v1", "a0", "a1", "a2", "a3",
    "t0", "t1", "t2", "t3", "t4", "t5", "t6", "t7", 
    "s0", "s1", "s2", "s3", "s4", "s5", "s6", "s7", 
    "t8", "t9", "k0", "k1", "gp", "sp", "s8", "ra"};

void Disassemble(char *code, WimpInst_t I, uint32_t nextpc)
{
    /** get the opcode field */
    switch(OPTYPE(I))
    {
        case RTYPE:
        {
            int rs, rt, rd, sa, func, x;
            rs   = RTYPE_RS(I);
            rt   = RTYPE_RT(I);
            rd   = RTYPE_RD(I);
            sa   = RTYPE_SA(I);
            func = RTYPE_FUNC(I);
            x    = RTYPE_X(I);

            switch(func)
            {
                case SLL:  if(I == 0)
                             sprintf(code, "nop"); 
                           else 
                             sprintf(code, "sll   %s,%s,%d", reg[rd], reg[rt], sa); return;
                case SRL:    sprintf(code, "srl   %s,%s,%d", reg[rd], reg[rt], sa); return;
                case SRA:    sprintf(code, "sra   %s,%s,%d", reg[rd], reg[rt], sa); return;
                case SLLV:   sprintf(code, "sllv  %s,%s,%s", reg[rd], reg[rt], reg[rs]); return;
                case SRLV:   sprintf(code, "srlv  %s,%s,%s", reg[rd], reg[rt], reg[rs]); return;
                case SRAV:   sprintf(code, "srav  %s,%s,%s", reg[rd], reg[rt], reg[rs]); return;
                case JR:     sprintf(code, "jr    %s      ", reg[rs]); return;
                case JALR:   sprintf(code, "jalr  %s,%s   ", reg[rd], reg[rs]); return;
                case SYSCALL:sprintf(code, "syscall %d", x); return;
                case BREAK:  sprintf(code, "break   %d", x); return;
                case MFHI:   sprintf(code, "mfhi  %s      ", reg[rd]); return;
                case MTHI:   sprintf(code, "mthi  %s      ", reg[rs]); return;
                case MFLO:   sprintf(code, "mflo  %s      ", reg[rd]); return;
                case MTLO:   sprintf(code, "mfto  %s      ", reg[rs]); return;
                case MULT:   sprintf(code, "mult  %s,%s   ", reg[rs], reg[rt]); return;
                case MULTU:  sprintf(code, "multu %s,%s   ", reg[rs], reg[rt]); return;
                case DIV:    sprintf(code, "div   %s,%s   ", reg[rs], reg[rt]); return;
                case DIVU:   sprintf(code, "divu  %s,%s   ", reg[rs], reg[rt]); return;
                case ADD:    sprintf(code, "add   %s,%s,%s", reg[rd], reg[rs], reg[rt]); return;
                case ADDU:   sprintf(code, "addu  %s,%s,%s", reg[rd], reg[rs], reg[rt]); return;
                case SUB:    sprintf(code, "sub   %s,%s,%s", reg[rd], reg[rs], reg[rt]); return;
                case SUBU:   sprintf(code, "subu  %s,%s,%s", reg[rd], reg[rs], reg[rt]); return;
                case AND:    sprintf(code, "and   %s,%s,%s", reg[rd], reg[rs], reg[rt]); return;
                case OR:     sprintf(code, "or    %s,%s,%s", reg[rd], reg[rs], reg[rt]); return;
                case XOR:    sprintf(code, "xor   %s,%s,%s", reg[rd], reg[rs], reg[rt]); return;
                case NOR:    sprintf(code, "nor   %s,%s,%s", reg[rd], reg[rs], reg[rt]); return;
                case SLT:    sprintf(code, "slt   %s,%s,%s", reg[rt], reg[rs], reg[rt]); return;
                case SLTU:   sprintf(code, "sltu  %s,%s,%s", reg[rd], reg[rs], reg[rt]); return;
               default:      sprintf(code, "dc 0x%x", I); return;
            }
        }
        case BRANCH:
        {
            enum condcode cond = ITYPE_COND(I);
            int rs = ITYPE_RS(I);
            int jmpdest = nextpc + SIGN_XTND(ITYPE_IMM(I)) * 4;
            switch(cond)
            {
                case LTZ:  sprintf(code, "bltz  %s,0x%08x", reg[rs], jmpdest); return;
                case GEZ:  sprintf(code, "bgez  %s,0x%08x", reg[rs], jmpdest); return;
                case LTZAL:sprintf(code, "bltzal %s,0x%08x", reg[rs], jmpdest); return;
                case GEZAL:sprintf(code, "bgezal %s,0x%08x", reg[rs], jmpdest); return;
                default:   sprintf(code, "dc 0x%x", I); return;
            }
        }
        case JUMP:  sprintf(code, "j     0x%08x", (nextpc & (0xf << 28)) | (JTYPE_TARGET(I) << 2)); return;
        case JAL:   sprintf(code, "jal   0x%08x", (nextpc & (0xf << 28)) | (JTYPE_TARGET(I) << 2)); return;
        case BEQ:   sprintf(code, "beq   %s,%s,0x%08x", reg[ITYPE_RS(I)], reg[ITYPE_RT(I)], nextpc + SIGN_XTND(ITYPE_IMM(I)) * 4); return;
        case BNE:   sprintf(code, "bne   %s,%s,0x%08x", reg[ITYPE_RS(I)], reg[ITYPE_RT(I)], nextpc + SIGN_XTND(ITYPE_IMM(I)) * 4); return;
        case BLEZ:  sprintf(code, "blez  %s,%s,0x%08x", reg[ITYPE_RS(I)], reg[ITYPE_RT(I)], nextpc + SIGN_XTND(ITYPE_IMM(I)) * 4); return;
        case BGTZ:  sprintf(code, "bgtz  %s,%s,0x%08x", reg[ITYPE_RS(I)], reg[ITYPE_RT(I)], nextpc + SIGN_XTND(ITYPE_IMM(I)) * 4); return;
        case ADDI:  sprintf(code, "addi  %s,%s,%d", reg[ITYPE_RT(I)], reg[ITYPE_RS(I)], SIGN_XTND(ITYPE_IMM(I))); return;
        case ADDIU: sprintf(code, "addiu %s,%s,%d", reg[ITYPE_RT(I)], reg[ITYPE_RS(I)], SIGN_XTND(ITYPE_IMM(I))); return;
        case SLTI:  sprintf(code, "slti  %s,%s,%d", reg[ITYPE_RT(I)], reg[ITYPE_RS(I)], SIGN_XTND(ITYPE_IMM(I))); return;
        case SLTIU: sprintf(code, "sltiu %s,%s,%d", reg[ITYPE_RT(I)], reg[ITYPE_RS(I)], SIGN_XTND(ITYPE_IMM(I))); return;
        case ANDI:  sprintf(code, "andi  %s,%s,0x%04x", reg[ITYPE_RT(I)], reg[ITYPE_RS(I)], ITYPE_IMM(I)); return;
        case ORI:   sprintf(code, "ori   %s,%s,0x%04x", reg[ITYPE_RT(I)], reg[ITYPE_RS(I)], ITYPE_IMM(I)); return;
        case XORI:  sprintf(code, "xori  %s,%s,0x%04x", reg[ITYPE_RT(I)], reg[ITYPE_RS(I)], ITYPE_IMM(I)); return;
        case LUI:   sprintf(code, "lui   %s,0x%04x",    reg[ITYPE_RT(I)], ITYPE_IMM(I)); return;
        case LW:    sprintf(code, "lw    %s,%d(%s)",    reg[ITYPE_RT(I)], SIGN_XTND(ITYPE_IMM(I)), reg[ITYPE_RS(I)]); return;
        case SW:    sprintf(code, "sw    %s,%d(%s)",    reg[ITYPE_RT(I)], SIGN_XTND(ITYPE_IMM(I)), reg[ITYPE_RS(I)]); return;
    }
}
