/**
 ** WIMP -- WIesbaden MIps Processor
 **
 ** WIMP instruction set
 **
 **
 ** Copyright (c) 2014, Robert Kaiser, RheinMain University of Applied Sciences
 ** All rights reserved.
 ** 
 ** Redistribution and use in source and binary forms, with or without modification,
 ** are permitted provided that the following conditions are met:
 **
 ** 1. Redistributions of source code must retain the above copyright notice, this
 **    list of conditions and the following disclaimer.
 ** 2. Redistributions in binary form must reproduce the above copyright notice,
 **    this list of conditions and the following disclaimer in the documentation
 **    and/or other materials provided with the distribution.
 ** 
 ** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 ** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 ** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 ** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 ** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 ** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 ** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 ** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 ** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 ** OF SUCH DAMAGE.
 **
 */

#ifndef __WIMP_OPCODES_H__
#define __WIMP_OPCODES_H__ 1

#include <stdint.h>

enum opcode
{
    RTYPE   = 0, /**< Register type instruction   */
    BRANCH  = 1, /**< PC-relative text and branch */
    JUMP    = 2, /**< unconditional 26-bit jump   */
    JAL     = 3, /**< 26-bit jump-and-link        */
    BEQ     = 4, /**< Jump on equal               */
    BNE     = 5, /**< Jump on not equal           */
    BLEZ    = 6, /**< Jump on less or equal zero  */
    BGTZ    = 7, /**< Jump on greater than zero   */
    ADDI    = 8, /**< Add immediate               */
    ADDIU   = 9, /**< Add immediate unsigned      */
    SLTI    = 10,/**< Set on less than immediate  */
    SLTIU   = 11,/**< Set on less than immediate unsigned */
    ANDI    = 12,/**< AND immediate               */
    ORI     = 13,/**< OR immediate                */
    XORI    = 14,/**< XOR immediate               */
    LUI     = 15,/**< load upper immediate        */
    /** coprocessor instrs omitted for now ...    */
    LB      = 32, /**< Load Byte                  */
    LH      = 33, /**< Load halfword and sign-extend */
    LWL     = 34, /**< Load word left             */
    LW      = 35, /**< Load word                  */
    LBU     = 36, /**< Load Byte and zero-extend  */
    LHU     = 37, /**< Load halfword and zero-extend */
    LWR     = 38, /**< Load word right            */

    SB      = 40, /**< Store byte                 */
    SH      = 41, /**< Store halfword             */
    SWL     = 42, /**< Store word left            */
    SW      = 43, /**< Store word                 */
    SWR     = 46, /**< Store word right           */
    NUM_INSTRUCTIONS = 47,
    PSEUDO  = 1000, /** pseudo opcodes            */
	SYNTH   = 2000, /** synthesized instructions  */
};

enum funcode {
    SLL     = 0,
    SRL     = 2,
    SRA     = 3,
    SLLV    = 4,
    SRLV    = 6,
    SRAV    = 7,
    JR      = 8,
    JALR    = 9,

    SYSCALL = 12,
    BREAK   = 13,
    MFHI    = 16,
    MTHI    = 17,
    MFLO    = 18,
    MTLO    = 19,
    MULT    = 24,
    MULTU   = 25,
    DIV     = 26,
    DIVU    = 27,

    ADD     = 32,
    ADDU    = 33,
    SUB     = 34,
    SUBU    = 35,
    AND     = 36,
    OR      = 37,
    XOR     = 38,
    NOR     = 39,
    SLT     = 42,
    SLTU    = 43,
	NUM_RTYPEFUNCS = 44,
};

enum condcode {
    LTZ    = 0,
    GEZ    = 1,
    LTZAL  = 16,
    GEZAL  = 17,
};


/**
 * @brief Get the opcode field of an instruction
 */
#define OPTYPE(instr) (((instr) >> 26) & 0x3f)


/**
 * @brief I-Type instruction structure
 */
#define ITYPE(op, rs, rt, imm) \
     ((((op) & 0x3f) << 26) | /**< The opcode           */ \
      (((rs) & 0x1f) << 21) | /**< Source register      */ \
      (((rt) & 0x1f) << 16) | /**< Target register      */ \
      (((imm) & 0xffff))    ) /**< immediate            */

/**
 * @brief Get various fields of an I-type instruction
 */
#define ITYPE_RS(instr) (((instr) >> 21) & 0x1f)
#define ITYPE_RT(instr) (((instr) >> 16) & 0x1f)
#define ITYPE_COND(instr) (((instr) >> 16) & 0x1f)
#define ITYPE_IMM(instr) ((instr) & 0xffff)


/**
 * @brief J-Type instruction structure
 */
#define JTYPE(op, target) \
     ((((op) & 0x3f) << 26)   |  /**< The opcode       */ \
      (((target) & 0x3ffffff)))  /**< Jump destination */

/**
 * @brief Get the target field of a J-type instruction
 */
#define JTYPE_TARGET(instr) ((instr) & 0x3ffffff)


/**
 * @brief R-Type instruction structure
 */
#define RTYPE(op, rs, rt, rd, sa, func) \
     ((((op) & 0x3f) << 26) | /**< The opcode          */ \
      (((rs) & 0x1f) << 21) | /**< Source register     */ \
      (((rt) & 0x1f) << 16) | /**< Target register     */ \
      (((rd) & 0x1f) << 11) | /**< Destination register*/ \
      (((sa) & 0x1f) <<  6) | /**< Shift amount        */ \
      (((func) & 0x3f))    )  /**< function field      */

/**
 * @brief Get various fields of an I-type instruction
 */
#define RTYPE_RS(instr) (((instr) >> 21) & 0x1f)
#define RTYPE_RT(instr) (((instr) >> 16) & 0x1f)
#define RTYPE_RD(instr) (((instr) >> 11) & 0x1f)
#define RTYPE_SA(instr) (((instr) >>  6) & 0x1f)
#define RTYPE_FUNC(instr) ((instr) & 0x3f)
#define RTYPE_X(instr) (((instr) >>  6) & 0xfffff)

/**
 * @brief Special R-Type instruction case: syscall od break instruction
 */
#define RTRAP(op, x, func) \
     ((((op) & 0x3f)  << 26) | /**< The opcode          */ \
      (((x) & 0xfffff) << 6) | /**< reserved field      */ \
      (((func) & 0x3f))      ) /**< function field      */


typedef uint32_t WimpInst_t;

#endif /* __WIMP_OPCODES_H__ */


