/**
 ** WIMP -- WIesbaden MIps Processor
 **
 ** WIMP Memory interface
 **
 ** Copyright (c) 2014, Robert Kaiser, RheinMain University of Applied Sciences
 ** All rights reserved.
 ** 
 ** Redistribution and use in source and binary forms, with or without modification,
 ** are permitted provided that the following conditions are met:
 **
 ** 1. Redistributions of source code must retain the above copyright notice, this
 **    list of conditions and the following disclaimer.
 ** 2. Redistributions in binary form must reproduce the above copyright notice,
 **    this list of conditions and the following disclaimer in the documentation
 **    and/or other materials provided with the distribution.
 ** 
 ** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 ** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 ** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 ** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 ** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 ** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 ** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 ** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 ** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 ** OF SUCH DAMAGE.
 **
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "wimp-mem.h"
#include "wimp-io.h"

/**
 * (mis-)use net functions for optional byte swapping:
 */
#include <netinet/in.h>
#define swapl(x) htonl(x)

extern FILE* tracefile;

/**
 ** Memory Image:
 */
static uint32_t *TheMemory;
static int MemorySize;
static mmio_t *iodevs = 0;

/** check if address is valid in data memory */
int M_DataAccessWord(int Address)
{
    if(Address >= 0 && Address < MemorySize && (Address & 3) == 0)
        return 1;
    else if(iodevs)
    {
		mmio_t *iodev;
	
		for(iodev = iodevs; iodev; iodev = iodev->next)
		{
			if(Address >= iodev->base && Address < iodev->base + iodev->size)
				return 1;
		}
	}
	return 0;
}

/** read a word from data memory */
uint32_t M_ReadDataWord(int Address)
{
    if((Address < 0 || Address >= MemorySize) && iodevs)
    {
		mmio_t *iodev;
	
		for(iodev = iodevs; iodev; iodev = iodev->next)
		{
			if(Address >= iodev->base && Address < iodev->base + iodev->size)
			{
				return iodev->read(iodev->private, Address - iodev->base);
			}
		}
    }
    if(tracefile)
        fprintf(tracefile, "0 %x\n", Address);
    return swapl(TheMemory[Address / sizeof(int)]);
}

/** write a word to data memory */
void M_WriteDataWord(unsigned int Address, uint32_t Data)
{
    if((Address >= MemorySize) && iodevs)
    {
		mmio_t *iodev;
	
		for(iodev = iodevs; iodev; iodev = iodev->next)
		{
			if(Address >= iodev->base && Address < iodev->base + iodev->size)
			{
                iodev->write(iodev->private, Address - iodev->base, Data);
				return;
			}
		}
    }
    if(tracefile)
        fprintf(tracefile, "1 %x\n", Address);
    TheMemory[Address / sizeof(uint32_t)] = swapl(Data);
}

/** Check if address is valid in program memory */
int M_ProgramAccessWord(unsigned int Address)
{
    return(Address < MemorySize && (Address & 3) == 0);
}

/** read a word from program memory */
uint32_t M_ReadProgramWord(unsigned int Address)
{
    if(tracefile)
        fprintf(tracefile, "2 %x\n", Address);
    return swapl(TheMemory[Address / sizeof(uint32_t)]);
}

/** create a new address space */
int M_CreateSpace(char *filename, unsigned int size)
{
    int fd;
    if((fd = open(filename, O_RDONLY)) < 0)
    {
        perror("open");
        return -1;
    }
    TheMemory = (uint32_t*)malloc(size + 4);
    MemorySize = size;

    if(read(fd, TheMemory, size) < size)
    {
        perror("read");
        return -1;
    }
    close(fd);
    return 0;
}

/** add memory-mapped I/O device to address space */
int M_AddMMIO(int ID, mmio_t *dev)
{
	mmio_t *iodev = iodevs;
	
	if(!iodev)
	{
		iodevs = dev;
		dev->next = NULL;
	}
	else
	{
		while(iodev->next)
			iodev = iodev->next;
		iodev->next = dev;
		dev->next = NULL;
	}
	return 0;
}


/** delete an address space */
void M_DeleteSpace(int ID)
{
    free(TheMemory);
    TheMemory = NULL;
}
