/**
 ** WIMP -- WIesbaden MIps Processor
 **
 ** WImp Executor
 **
 ** Usage: wie <program.wo>
 **
 ** Copyright (c) 2014, Robert Kaiser, RheinMain University of Applied Sciences
 ** All rights reserved.
 ** 
 ** Redistribution and use in source and binary forms, with or without modification,
 ** are permitted provided that the following conditions are met:
 **
 ** 1. Redistributions of source code must retain the above copyright notice, this
 **    list of conditions and the following disclaimer.
 ** 2. Redistributions in binary form must reproduce the above copyright notice,
 **    this list of conditions and the following disclaimer in the documentation
 **    and/or other materials provided with the distribution.
 ** 
 ** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 ** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 ** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 ** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 ** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 ** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 ** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 ** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 ** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 ** OF SUCH DAMAGE.
 **
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>

#include "wimp-opcodes.h"
#include "wimp-mem.h"
#include "wimp-io.h"
#include "wimp-dis.h"


#define NUM_REGISTERS  32

#define SIGN_XTND(x) ((int32_t)((int16_t)(x)))


int  verbose    = 0;
int  singlestep = 0;
FILE *tracefile = NULL;

static struct ProcessorState
{
    uint32_t     Register[NUM_REGISTERS];
    uint32_t     Lo, Hi;
    unsigned int PC;
    int          Halted;
} Processor;


static void dump_CPUstate(void)
{
    int i;
    char disassembly[64];
    static char *regname[32] = {
        "zero", "at", "v0", "v1", "a0", "a1", "a2", "a3",
        "  t0", "t1", "t2", "t3", "t4", "t5", "t6", "t7", 
        "  s0", "s1", "s2", "s3", "s4", "s5", "s6", "s7", 
        "  t8", "t9", "k0", "k1", "gp", "sp", "s8", "ra"
    };
    for(i = 0; i < 32; i++)
    {
        if((i % 8) == 0)
            printf("\n");
        printf("%s=%08x ", regname[i], Processor.Register[i]);
    }
    if(M_ProgramAccessWord(Processor.PC))
    {
        Disassemble(disassembly, M_ReadProgramWord(Processor.PC), Processor.PC);
        printf("\n%08x:%s >", Processor.PC, disassembly);
    }
    else
    {
        printf("\n%08x:???????? >", Processor.PC);
    }
    fflush(stdout);
}



/** Thow an exception: for now this just terminates the processor */
static void exception(char* text)
{
    printf("******* Exception: %s *******\n", text);
    if(verbose)
            dump_CPUstate();
    exit(1);
}



static void I_invalid(WimpInst_t I)
{
     exception("invalid opcode!");
}

static void I_sll(int rs, int rt, int rd, int sa, int x)
{
    Processor.Register[rd] = Processor.Register[rt] << sa;
}

static void I_srl(int rs, int rt, int rd, int sa, int x)
{
    assert(rs == 0);
    Processor.Register[rd] = Processor.Register[rt] >> sa;
}

static void I_sra(int rs, int rt, int rd, int sa, int x)
{
    assert(rs == 0);
    Processor.Register[rd] = (Processor.Register[rt] >> sa) | (0xffffffff << (32 - sa));
}

static void I_sllv(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    sa = Processor.Register[rs] & 0x1f;
    Processor.Register[rd] = Processor.Register[rt] << sa;
}

static void I_srlv(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    sa = Processor.Register[rs] & 0x1f;
    Processor.Register[rd] = Processor.Register[rt] >> sa;
}

static void I_srav(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    sa = Processor.Register[rt] & 0x1f;
    Processor.Register[rd] = (Processor.Register[rt] >> sa) | (0xffffffff << (32 - sa));
}

static void I_jr(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    assert(rd == 0);
    assert(rt == 0);
    if((Processor.Register[rs] & 0x3))
        exception("Unaligned target address (JR)!");
    Processor.PC = Processor.Register[rs];
}

static void I_jalr(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    assert(rd == 0);
    if((Processor.Register[rs] & 0x3))
        exception("Unaligned target address (JALR)!");
    Processor.Register[rd] = Processor.PC;
    Processor.PC = Processor.Register[rs];
}

static void I_syscall(int rs, int rt, int rd, int sa, int x)
{
	static int new_line = 1;
	switch(x)
    {
        case 0:
            Processor.Halted = 1;
            break;
        case 1: /** syscall 1: print contents of $a0 as decimal number */
            if(verbose > 0)
                printf("Outdec: %d\n", Processor.Register[4]);
            else
                printf("%d", Processor.Register[4]);
            break;
        case 2: /** syscall 2: print contents of $a0 as hex number */
            if(verbose > 0)
                printf("Outhex: %x\n", Processor.Register[4]);
            else
                printf("%x", Processor.Register[4]);
            break;
        case 3: /** syscall 3: print contents of $a0 as character */
        {   /** print prompt only if beginning a new line */
            char c = Processor.Register[4] & 0xff;
            if(new_line)
            {   /** print prompt only if a new line has been started */
                if(verbose > 0)
                    printf("Outc: %c", c);
                else
                    putchar(c);
                new_line = 0;
            }
            else
            {
                putchar(c);
                if(c == '\n')
                    new_line = 1;
            }
            break;
        }
        case 4: /** syscall 4: input a number, deliver in $v0 */
            if(verbose > 0)
                printf("Inputdec>");
            fflush(stdout);
            if(scanf("%d", &Processor.Register[2])!= 1)
                fflush(stdin);
            break;
        case 5: /** syscall 5: input a number, deliver in $v0 */
            if(verbose > 0)
                printf("Inputhex>");
            fflush(stdout);
            if(scanf("%x", &Processor.Register[2])!= 1)
                fflush(stdin);
            break;
        case 6: /** syscall 6: input a character, deliver in $v0 */
        {   /** print prompt only if beginning a new line */
            if(new_line)
            {
                if(verbose > 0)
                {
                    printf("Input>");
                    fflush(stdout);
                    new_line = 0;
                }
                if((Processor.Register[2] = getchar()) == '\n')
                {
                    Processor.Register[2] = '\0';
                    new_line = 1;
                }
            }
        }
        default:
            exception("Unhandled SYSCALL!");
            break;
    }
}

static void I_break(int rs, int rt, int rd, int sa, int x)
{
    if(verbose > 0)
    {
        printf("Break Instruction (%d) at 0x%08x", x, Processor.PC - 4);
        dump_CPUstate();
        if(verbose > 1)
            if(getchar() == 's')
                 singlestep++;
    }
}

static void I_mfhi(int rs, int rt, int rd, int sa, int x)
{
    assert(rs == 0);
    assert(rt == 0);
    assert(sa == 0);
    Processor.Register[rd] = Processor.Hi;
}

static void I_mthi(int rs, int rt, int rd, int sa, int x)
{
    assert(rd == 0);
    assert(rt == 0);
    assert(sa == 0);
    Processor.Hi = Processor.Register[rs];
}

static void I_mflo(int rs, int rt, int rd, int sa, int x)
{
    assert(rs == 0);
    assert(rt == 0);
    assert(sa == 0);
    Processor.Register[rd] = Processor.Lo;
}

static void I_mtlo(int rs, int rt, int rd, int sa, int x)
{
    assert(rd == 0);
    assert(rt == 0);
    assert(sa == 0);
    Processor.Lo = Processor.Register[rs];
}

static void I_mult(int rs, int rt, int rd, int sa, int x)
{
    uint64_t res;
    assert(rd == 0);
    assert(sa == 0);
    res = (int64_t)Processor.Register[rs] * (int64_t)Processor.Register[rt];
    Processor.Hi = res >> 32;
    Processor.Lo = res & 0xFFFFFFFFULL;
}

static void I_multu(int rs, int rt, int rd, int sa, int x)
{
    uint64_t res;
    assert(rd == 0);
    assert(sa == 0);
    res = (uint64_t)Processor.Register[rs] * (uint64_t)Processor.Register[rt];
    Processor.Hi = res >> 32;
    Processor.Lo = res & 0xFFFFFFFFULL;
}

static void I_div(int rs, int rt, int rd, int sa, int x)
{
    Processor.Lo = (int32_t)Processor.Register[rs] / (int32_t)Processor.Register[rt];
    Processor.Hi = (int32_t)Processor.Register[rs] % (int32_t)Processor.Register[rt];
}

static void I_divu(int rs, int rt, int rd, int sa, int x)
{
    Processor.Lo = (uint32_t)Processor.Register[rs] / (uint32_t)Processor.Register[rt];
    Processor.Hi = (uint32_t)Processor.Register[rs] % (uint32_t)Processor.Register[rt];
}

static void I_add(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    Processor.Register[rd] = Processor.Register[rs] + Processor.Register[rt];
    /** FIXME: Throw an exception on overflow */
}

static void I_addu(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    Processor.Register[rd] = Processor.Register[rs] + Processor.Register[rt];
}

static void I_sub(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    Processor.Register[rd] = Processor.Register[rs] - Processor.Register[rt];
    /** FIXME: Throw an exception on overflow */
}

static void I_subu(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    Processor.Register[rd] = Processor.Register[rs] - Processor.Register[rt];
}

static void I_and(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    Processor.Register[rd] = Processor.Register[rs] & Processor.Register[rt];
}

static void I_or(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    Processor.Register[rd] = Processor.Register[rs] | Processor.Register[rt];
}

static void I_xor(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    Processor.Register[rd] = Processor.Register[rs] ^ Processor.Register[rt];
}

static void I_nor(int rs, int rt, int rd, int sa, int x)
{
    assert(sa == 0);
    Processor.Register[rd] = ~(Processor.Register[rs] | Processor.Register[rt]);
}

static void I_slt(int rs, int rt, int rd, int sa, int x)
{
    if((int32_t)Processor.Register[rs] < (int32_t)Processor.Register[rt])
        Processor.Register[rd] = 1;
    else
        Processor.Register[rd] = 0;
}

static void I_sltu(int rs, int rt, int rd, int sa, int x)
{
    if((uint32_t)Processor.Register[rs] < (uint32_t)Processor.Register[rt])
        Processor.Register[rd] = 1;
    else
        Processor.Register[rd] = 0;
}

static void I_rtype_invalid(int rs, int rt, int rd, int sa, int x)
{
    assert(0); /** should never get here */
/*    I_invalid(RTYPE(RTYPE, rs, rt, rd, sa, -1));*/
}

/**
 * The R-type sub-instruction set array
 */

typedef void (*rtypeinstruction_t)(int rs, int rt, int rd, int sa, int x);

static rtypeinstruction_t RtypeInstructionSet[NUM_RTYPEFUNCS];

static void I_rtype(WimpInst_t I)
{
    int rs, rt, rd, sa, func, x;
    rtypeinstruction_t Exec;
    rs   = RTYPE_RS(I);
    rt   = RTYPE_RT(I);
    rd   = RTYPE_RD(I);
    sa   = RTYPE_SA(I);
    func = RTYPE_FUNC(I);
    x    = RTYPE_X(I);
    if(func < 0 || func >= NUM_RTYPEFUNCS ||
      ((Exec = RtypeInstructionSet[func]) == I_rtype_invalid))
    {
        I_invalid(RTYPE(RTYPE, rs, rt, rd, sa, func));
    }
    else
    {
        Exec(rs, rt, rd, sa, x);
    }
}

static void I_branch(WimpInst_t I)
{
    enum condcode cond = ITYPE_COND(I);
    int rs = ITYPE_RS(I);
    int jmpdist = SIGN_XTND(ITYPE_IMM(I)) * 4;
    switch(cond)
    {
        case LTZ:
            if((int32_t)Processor.Register[rs] < 0)
                Processor.PC += jmpdist;
            break;
        case GEZ:
            if((int32_t)Processor.Register[rs] >= 0)
                Processor.PC += jmpdist;
            break;
        case LTZAL:
            Processor.Register[31] = Processor.PC;
            if((int32_t)Processor.Register[rs] < 0)
                Processor.PC += jmpdist;
            break;
        case GEZAL:
            Processor.Register[31] = Processor.PC;
            if((int32_t)Processor.Register[rs] >= 0)
                Processor.PC += jmpdist;
            break;
        default:
            I_invalid(I);
            break;
    }
}

static void I_jump(WimpInst_t I)
{
    Processor.PC = (Processor.PC & (0xf << 28)) | (JTYPE_TARGET(I) << 2);
}

static void I_jal(WimpInst_t I)
{
    Processor.Register[31] = Processor.PC;
    Processor.PC = (Processor.PC & (0xf << 28)) | (JTYPE_TARGET(I) << 2);
}

static void I_beq(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int rt = ITYPE_RT(I);
    int jmpdist = SIGN_XTND(ITYPE_IMM(I)) * 4;
    if(Processor.Register[rs] == Processor.Register[rt])
        Processor.PC += jmpdist;
}

static void I_bne(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int rt = ITYPE_RT(I);
    int jmpdist = SIGN_XTND(ITYPE_IMM(I)) * 4;
    if(Processor.Register[rs] != Processor.Register[rt])
        Processor.PC += jmpdist;
}

static void I_blez(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int jmpdist = SIGN_XTND(ITYPE_IMM(I)) * 4;
    assert(ITYPE_RT(I) == 0);
    if(Processor.Register[rs] <= 0)
        Processor.PC += jmpdist;
}

static void I_bgtz(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int jmpdist = SIGN_XTND(ITYPE_IMM(I)) * 4;
    assert(ITYPE_RT(I) == 0);
    if(Processor.Register[rs] > 0)
        Processor.PC += jmpdist;
}

static void I_addi(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int rt = ITYPE_RT(I);
    int imm = SIGN_XTND(ITYPE_IMM(I));
    Processor.Register[rt] = Processor.Register[rs] + imm;
    /** FIXME: Throw an exception on overflow */
}

static void I_addiu(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int rt = ITYPE_RT(I);
    int imm = SIGN_XTND(ITYPE_IMM(I));
    Processor.Register[rt] = Processor.Register[rs] + imm;
}

static void I_slti(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int rt = ITYPE_RT(I);
    int imm = SIGN_XTND(ITYPE_IMM(I));
    
    if((int32_t)Processor.Register[rs] < imm)
        Processor.Register[rt] = 1;
    else
        Processor.Register[rt] = 0;
}

static void I_sltiu(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int rt = ITYPE_RT(I);
    int imm = SIGN_XTND(ITYPE_IMM(I));
    
    if((uint32_t)Processor.Register[rs] < (uint32_t)imm)
        Processor.Register[rt] = 1;
    else
        Processor.Register[rt] = 0;
}

static void I_andi(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int rt = ITYPE_RT(I);
    int imm = ITYPE_IMM(I);
    Processor.Register[rt] = Processor.Register[rs] & imm;
}

static void I_ori(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int rt = ITYPE_RT(I);
    int imm = ITYPE_IMM(I);
    Processor.Register[rt] = Processor.Register[rs] | imm;
}

static void I_xori(WimpInst_t I)
{
    int rs = ITYPE_RS(I);
    int rt = ITYPE_RT(I);
    int imm = ITYPE_IMM(I);
    Processor.Register[rt] = Processor.Register[rs] ^ imm;
}

static void I_lui(WimpInst_t I)
{
    int rt = ITYPE_RT(I);
    int imm = ITYPE_IMM(I);
    assert(ITYPE_RS(I) == 0);
    Processor.Register[rt] = imm << 16;
}

static void I_lw(WimpInst_t I)
{
     unsigned int addr = Processor.Register[ITYPE_RS(I)] + SIGN_XTND(ITYPE_IMM(I));
     if(!M_DataAccessWord(addr))
     {
          exception("LW from invalid address!");
     }
     Processor.Register[ITYPE_COND(I)] = M_ReadDataWord(addr);
}

static void I_sw(WimpInst_t I)
{
     unsigned int addr = Processor.Register[ITYPE_RS(I)] + SIGN_XTND(ITYPE_IMM(I));
     if(!M_DataAccessWord(addr))
     {
          exception("SW to invalid address!");
     }
     M_WriteDataWord(addr, Processor.Register[ITYPE_COND(I)]);
}


/**
 * The instruction set array
 */

typedef void (*instruction_t)(WimpInst_t I);

static instruction_t InstructionSet[NUM_INSTRUCTIONS];

static void InitInstructionSet(void)
{
    int i;
    /** Default-init invalid */
    for(i = 0; i < NUM_INSTRUCTIONS; i++)
    {
        InstructionSet[i] = I_invalid;
    }
    /** set supported instructions */
    InstructionSet[RTYPE] = I_rtype;
    InstructionSet[BRANCH]= I_branch;
    InstructionSet[JUMP]  = I_jump;
    InstructionSet[JAL]   = I_jal;
    InstructionSet[BEQ]   = I_beq;
    InstructionSet[BNE]   = I_bne;
    InstructionSet[BLEZ]  = I_blez;
    InstructionSet[BGTZ]  = I_bgtz;
    InstructionSet[ADDI]  = I_addi;
    InstructionSet[ADDIU] = I_addiu;
    InstructionSet[SLTI]  = I_slti;
    InstructionSet[SLTIU] = I_sltiu;
    InstructionSet[ANDI]  = I_andi;
    InstructionSet[ORI]   = I_ori;
    InstructionSet[XORI]  = I_xori;
    InstructionSet[LUI]   = I_lui;
    InstructionSet[LW]    = I_lw;
    InstructionSet[SW]    = I_sw;
    /** setup R-type instrucion set */
    for(i = 0; i < NUM_RTYPEFUNCS; i++)
    {
          RtypeInstructionSet[i] = I_rtype_invalid;
    }
    RtypeInstructionSet[SLL]     = I_sll;
    RtypeInstructionSet[SRL]     = I_srl;
    RtypeInstructionSet[SRA]     = I_sra;
    RtypeInstructionSet[SLLV]    = I_sllv;
    RtypeInstructionSet[SRLV]    = I_srlv;
    RtypeInstructionSet[SRAV]    = I_srav;
    RtypeInstructionSet[JR]     =  I_jr;
    RtypeInstructionSet[JALR]    = I_jalr;
    RtypeInstructionSet[SYSCALL] = I_syscall;
    RtypeInstructionSet[BREAK]   = I_break;
    RtypeInstructionSet[MFHI]    = I_mfhi;
    RtypeInstructionSet[MTHI]    = I_mthi;
    RtypeInstructionSet[MFLO]    = I_mflo;
    RtypeInstructionSet[MTLO]    = I_mtlo;
    RtypeInstructionSet[MULT]    = I_mult;
    RtypeInstructionSet[MULTU]   = I_multu;
    RtypeInstructionSet[DIV]     = I_div;
    RtypeInstructionSet[DIVU]    = I_divu;
    RtypeInstructionSet[ADD]     = I_add;
    RtypeInstructionSet[ADDU]    = I_addu;
    RtypeInstructionSet[SUB]     = I_sub;
    RtypeInstructionSet[SUBU]    = I_subu;
    RtypeInstructionSet[AND]     = I_and;
    RtypeInstructionSet[OR]      = I_or;
    RtypeInstructionSet[XOR]     = I_xor;
    RtypeInstructionSet[NOR]     = I_nor;
    RtypeInstructionSet[SLT]     = I_slt;
    RtypeInstructionSet[SLTU]    = I_sltu;
}



/** The Processor: */
static void TheProcessor()
{
    WimpInst_t Instruction;
    instruction_t Exec;
    int OpCode;

    InitInstructionSet();
    Processor.PC     = 0;
    Processor.Halted = 0;

    while(!Processor.Halted)
    {
        if(!M_ProgramAccessWord(Processor.PC))
        {
            printf("PC=%d\n", Processor.PC);
            exception("invalid memory address!");
        }

        /** fetch next instruction */
        Instruction = M_ReadProgramWord(Processor.PC);
        Processor.Register[0] = 0;

        if(singlestep)
        {
            dump_CPUstate();
            if(getchar() == 'c')
                singlestep = 0;;
        }
        Processor.PC += sizeof(WimpInst_t);

        /** get the opcode field */
        OpCode = OPTYPE(Instruction);

        if(OpCode < 0 || OpCode >= NUM_INSTRUCTIONS)
        {
            I_invalid(Instruction);
        }
        else
        {        /** Execute instruction */
            Exec = InstructionSet[OpCode];
            Exec(Instruction);
        }
    }
}

static void rtfm(char* argv[])
{
    printf("Usage: %s {Options} <program file>\n", argv[0]);
    printf("Options:\n");
    printf("    -v              be verbose\n");
	printf("    -s              start in single step mode\n");
	printf("    -t <trace file> write trace file\n");
}


int main(int argc, char *argv[])
{
    struct stat statbuf;
    int option;

    /**
     * Parse commandline options: see "man getopt"
     */
    while ((option = getopt(argc, argv, "?hvst:")) != -1)
    {
        switch(option)
        {
            case '?':
            case 'h':
                rtfm(argv);
                return EXIT_FAILURE;
                break;
            case 'v':
                ++verbose;
                break;
            case 's':
                ++singlestep;
                break;
            case 't':
                tracefile = fopen(optarg, "w");
                if (NULL == tracefile)
                {
                    perror("fopen"); 
                    return EXIT_FAILURE;
                }
                break;
        }
    }
    if(argc - optind != 1)
    {
        rtfm(argv);
        return EXIT_FAILURE;
    }
    if(stat(argv[optind], &statbuf) != 0)
    {
        perror("stat");
        return EXIT_FAILURE;
    }
    if(M_CreateSpace(argv[optind], statbuf.st_size) < 0)
        return  EXIT_FAILURE;
    if(install_uart())
        return  EXIT_FAILURE;
    if(install_uart2())
        return  EXIT_FAILURE;
    TheProcessor();
    /** Stop instruction: Halt processor */
    printf("******* Program ended *******\n");

    M_DeleteSpace(0);
    return EXIT_SUCCESS;
}

