/**
 ** WIMP -- WIesbaden MIps Processor
 **
 ** WIMP UART simulation
 **
 ** Copyright (c) 2014, Robert Kaiser, RheinMain University of Applied Sciences
 ** All rights reserved.
 ** 
 ** Redistribution and use in source and binary forms, with or without modification,
 ** are permitted provided that the following conditions are met:
 **
 ** 1. Redistributions of source code must retain the above copyright notice, this
 **    list of conditions and the following disclaimer.
 ** 2. Redistributions in binary form must reproduce the above copyright notice,
 **    this list of conditions and the following disclaimer in the documentation
 **    and/or other materials provided with the distribution.
 ** 
 ** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 ** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 ** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 ** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 ** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 ** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 ** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 ** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 ** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 ** OF SUCH DAMAGE.
 **
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <stddef.h>

#include "wimp-mem.h"
#include "wimp-io.h"


extern int verbose;


#define UART_BASE 0xffffff00

typedef struct
{
	volatile uint32_t rx;   /* read-only: receive data    (bits 0-7; rest: undefined) */
	volatile uint32_t stat; /* read-only: status of UART */
	volatile uint32_t tx;   /* write-only: data to transmit (bits 0-7; rest: reserved, write as 0) */
	volatile uint32_t ctrl; /* write only: UART configuration (currently no effect) */
} uart_t;

/* bits in stat register: */
#define STAT_RXFULL   (1<<0)   /** Received data ready (RX is full) */
#define STAT_TXEMPTY  (1<<1)   /** Ready to transmit (TX is empty) */



static uint32_t    uart_read(void *dev, unsigned int offset);
static void        uart_write(void *dev, unsigned int offset, uint32_t Data);

static struct uart_private
{
	int uart_ctrl;
} uart_dev;


static mmio_t uart_mmio = {
    NULL,
    UART_BASE,
    sizeof(uart_t),
    uart_read,
    uart_write,
    &uart_dev,
};

int install_uart(void)
{
	return M_AddMMIO(0, &uart_mmio);
}


static uint32_t uart_read(void *dev, unsigned int offset)
{
	switch(offset)
	{
		case offsetof(uart_t, rx):	/** reading from receive register */
		    return getchar();
		case offsetof(uart_t, stat):/** reading from status register */
		    return STAT_RXFULL|STAT_TXEMPTY; /** simulated UART is always ready */
		default:
		    if(verbose)
		        printf("Error: reading from invalid UART register %d\n", offset);
		    return 0;
	}
}

static void uart_write(void *dev, unsigned int offset, uint32_t Data)
{
	struct uart_private *device = (struct uart_private *)dev;
	switch(offset)
	{
		case offsetof(uart_t, tx):	/** writing to transmit register */
		    putchar(Data);
		    break;
		case offsetof(uart_t, ctrl):/** reading from status register */
		    device->uart_ctrl = Data;  /** just store the value (no function yet) */
		    break;
		default:
		    if(verbose)
		        printf("Error: writing to invalid UART register %d\n", offset);
		    break;
	}
}





