/**
 ** WIMP -- WIesbaden MIps Processor
 **
 ** WIMP UART simulation
 **
 ** Copyright (c) 2014, Robert Kaiser, RheinMain University of Applied Sciences
 ** All rights reserved.
 ** 
 ** Redistribution and use in source and binary forms, with or without modification,
 ** are permitted provided that the following conditions are met:
 **
 ** 1. Redistributions of source code must retain the above copyright notice, this
 **    list of conditions and the following disclaimer.
 ** 2. Redistributions in binary form must reproduce the above copyright notice,
 **    this list of conditions and the following disclaimer in the documentation
 **    and/or other materials provided with the distribution.
 ** 
 ** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 ** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 ** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 ** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 ** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 ** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 ** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 ** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 ** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 ** OF SUCH DAMAGE.
 **
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <stddef.h>

#include "wimp-mem.h"
#include "wimp-io.h"


extern int verbose;


#define UART_BASE 0x8000

typedef struct
{
	volatile uint32_t tx;       /* write-only: data to transmit (bits 0-7; rest: reserved, write as 0) */
	volatile uint32_t tx_start; /* write-only: start transmtting */
	volatile uint32_t tx_ready; /* read-only: status of UART */
} uart_t;

/* bits in start register: */
#define STAT_TXSTART  (1<<0)   /** Begin transmitting */
#define STAT_TXBUSY   (1<<0)   /** TX is busy         */



static uint32_t    uart_read(void *dev, unsigned int offset);
static void        uart_write(void *dev, unsigned int offset, uint32_t Data);

static struct uart_private
{
	int uart_tx;
	int uart_ctrl;
	int uart_stat;
} uart_dev;


static mmio_t uart_mmio = {
    NULL,
    UART_BASE,
    sizeof(uart_t),
    uart_read,
    uart_write,
    &uart_dev,
};

int install_uart2(void)
{
	uart_dev.uart_stat = 0;
	return M_AddMMIO(0, &uart_mmio);
}


static uint32_t uart_read(void *dev, unsigned int offset)
{
	uint32_t stat;
	struct uart_private *device = (struct uart_private *)dev;
	switch(offset)
	{
		case offsetof(uart_t, tx_ready):	/** reading from tx ready register */
			stat = device->uart_stat;
			device->uart_ctrl &= ~STAT_TXSTART;
			device->uart_stat &= ~STAT_TXBUSY;
			return stat;                    /** simulated UART is always ready */
		default:
		    if(verbose)
		        printf("Error: reading from invalid UART register %d\n", offset);
		    return 0;
	}
}

static void uart_write(void *dev, unsigned int offset, uint32_t Data)
{
	struct uart_private *device = (struct uart_private *)dev;
	switch(offset)
	{
		case offsetof(uart_t, tx):	/** writing to transmit register */
			device->uart_tx = Data; /** write to "register" */
		    break;
		case offsetof(uart_t, tx_start):/** writing tx start register */
		    device->uart_ctrl = Data;
			if(device->uart_ctrl & STAT_TXSTART)
			{   /** starting TX: output whatever is in the TX "register" */
				putchar(device->uart_tx);
				/** indicate transmit is busy */
				device->uart_stat |= STAT_TXBUSY;
			}
		    break;
		default:
		    if(verbose)
		        printf("Error: writing to invalid UART register %d\n", offset);
		    break;
	}
}





